/*
------------------------------------------------------------
	filemap.C - Simple Portable File Path Mapper
------------------------------------------------------------
 *  David Olofson, 2001
 * This code is released under the terms of the GNU LGPL.
 */

#include "config.h"

#include "filemap.h"

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>

#ifdef HAVE_STAT
#include <sys/stat.h>
#include <unistd.h>
#endif

/*
 * Delete the 'count' first bytes of a string.
 * Does *not* realloc(), so it's safe to use
 * on static or local (stack) buffers.
 *
 * Returns new length of string.
 */
static int strdel(char *str, int count)
{
	int len = strlen(str);
	memmove(str, str+count, len-count+1);
	return len-count;
}


/*
 * Inserts string 'ins' into 'str', starting at 'pos'.
 * Does *not* realloc(), so it's safe to use
 * on static or local (stack) buffers.
 *
 * Relies on FM_BUFFER_SIZE to prevent buffer overflow!
 * (In fact, it writes to (str + FM_BUFFER_SIZE - 1) to
 * ensure that the result is NULL terminated.)
 *
 * Returns new length of string.
 */
static int strins(char *str, const char *ins, int pos)
{
	char *cpy = strdup(str+pos);
	if(!cpy)
		return strlen(str);	//Silly but harmless...

	snprintf(str+pos, FM_BUFFER_SIZE-pos-1, "%s%s", ins, cpy);
	str[FM_BUFFER_SIZE-1] = 0;
	free(cpy);
	return strlen(str);
}

#if 0
using std::fopen;
using std::opendir;
using std::mkdir;
#endif

filemapper_t::filemapper_t()
{
	keys = NULL;
	app_path = strdup("./");
	next_buffer = 0;

	/*
	 * I'll put this here for now.
	 * Don't know if it's the perfect place...
	 */
	char *home = (char *)getenv("HOME");
	if(home)
		addpath("HOME", home);
}


filemapper_t::~filemapper_t()
{
	fm_key_t *k = keys;
	while(k)
	{
		fm_key_t *nk = k->next;
		free(k->path);
		free(k);
		k = nk;
	}

	free(app_path);
}


char *filemapper_t::salloc()
{
	char *b = buffers[next_buffer++];
	if(next_buffer >= FM_BUFFERS)
		next_buffer = 0;
	return b;
}


void filemapper_t::no_double_slashes(char *buf)
{
	/*
	 * Clean slashes
	 */
	char *c = buf;
	while(*c)
	{
		if(*c == '/')
			while(c[1] == '/')
				strdel(c+1, 1);
		++c;
	}
}


void filemapper_t::unix_slashes(char *buf)
{
	/*
	 * Convert "~/" into "HOME>>".
	 */
	if(strncmp(buf, "~/", 2) == 0)
	{
		strdel(buf, 2);
		strins(buf, "HOME"FM_DEREF_TOKEN, 0);
	}

#if defined(WIN32)
	if(strchr(buf, ':'))
		strins(buf, FM_DEREF_TOKEN, 0);

	int len = strlen(buf);
	for(int i = 0; i < len; ++i)
		if(buf[i] == '\\')
			buf[i] = '/';
#elif defined(MACOS)
	int start = 0;
	if(strncmp(buf, FM_DEREF_TOKEN, 2) == 0)
		start = 2;

	int len = strlen(buf);
	for(int i = start; i < len; ++i)
		if(buf[i] == ':')
			buf[i] = '/';
#endif
	no_double_slashes(buf);
}


void filemapper_t::sys_slashes(char *buf)
{
	no_double_slashes(buf);

#if defined(WIN32)
	int len = strlen(buf);
	for(int i = 0; i < len; ++i)
		if(buf[i] == '/')
			buf[i] = '\\';
#elif defined(MACOS)
	/* 
	 * Untested.
	 */
	int len;
	if(strncmp(buf, "./", 2) == 0)
		len = strdel(buf, 2);
	else
		len = strlen(buf);
	/* Fix absolute/relative */
	if(buf[0] == '/')
		len = strdel(buf, 1);
	else
	{
		memmove(buf + 1, buf, len);
		++len;
		buf[0] = ':';
	}
	/* 
	 * '/' --> ':'
	 */
	for(int i = 0; i < len; ++i)
		if(buf[i] == '/')
			buf[i] = ':';
	/* 
	 * "..:" --> "::"
	 */
	for(int i = 0; i < len; ++i)
		if(strncmp(buf + i, "..:", 3) == 0)
		{
			strdel(buf + i, 1);
			buf[i] = ':';
			--len;
		}
#endif
}


void filemapper_t::exepath(const char *appname)
{
	if(app_path)
		free(app_path);

	try
	{
		app_path = strdup(appname);
	}
	catch(...)
	{
#if defined(WIN32)
		app_path = strdup("");
#elif defined(MACOS)
		app_path = strdup(":");
#else
		app_path = strdup("./");
#endif
		addpath("EXE", app_path);
		return;
	}

#if defined(WIN32)
	char *c = strrchr(app_path, '\\');
	if(c)
		c[1] = 0;
	else
	{
		free(app_path);
		app_path = strdup("");
	}
#elif defined(MACOS)
	char *c = strrchr(app_path, ':');
	if(c)
		c[1] = 0;
	else
	{
		free(app_path);
		app_path = strdup(":");
	}
#else
	char *c = strrchr(app_path, '/');
	if(c)
		c[1] = 0;
	else
	{
		free(app_path);
		app_path = strdup("./");
	}
#endif
	addpath("EXE", app_path);
	DBG(printf("Application path: '%s'\n", app_path);)
}


void filemapper_t::addpath(const char *key, const char *path, int first)
{
	fm_key_t *k, *insk;
	try
	{
		k = new fm_key_t;
		strncpy(k->key, key, 8);
		k->key[8] = 0;
		k->path = strdup(path);
		if(first)
		{
			k->next = keys;
			keys = k;
		}
		else
		{
			k->next = NULL;
			if(keys)
			{
				insk = keys;
				while(insk->next)
					insk = insk->next;
				insk->next = k;
			}
			else
				keys = k;
		}
	}
	catch(...)
	{
		return;
	}
}


fm_key_t *filemapper_t::getkey(fm_key_t *key = NULL, const char *ref = NULL)
{
	int all = ref ? 0 : (ref[0] == '*');
	if(key)
	{
		if(!ref)
			ref = key->key;
		else if(all)
			return key->next;
		else
			key = key->next;
	}
	else
	{
		if(!ref || all)
			return keys;
		else
			key = keys;
	}

	while(key)
	{
		if(strcmp(key->key, ref) == 0)
			return key;
		key = key->next;
	}
	return NULL;
}


/*
 * Checks if an object exists, and if so, what kind it is.
 */
int filemapper_t::probe(char *syspath)
{
#ifdef HAVE_STAT
	struct stat st;
	if(stat(syspath, &st) == 0)
	{
		if(S_ISDIR(st.st_mode))
			return FM_DIR;
		else
			return FM_FILE;
	}
#else
	/*
	 * On Linux, fopen(..., "r") seems to always
	 * succeed, even on directories. Weird...
	 */
	FILE *f = ::fopen(syspath, "r+");
	if(f)
	{
		fclose(f);
		return FM_FILE;
	}
	else
		if(errno == EISDIR)
			return FM_DIR;
#endif
	return FM_ERROR;
}


/*
 * Create a file for writing, or if it exists, test if it
 * can be opened in write mode.
 *
 * Returns
 *	FM_FILE if the file exists and is writable,
 *	FM_FILE_CREATE if the file had to be created,
 *	FM_DIR if the path leads to a directory, or
 *	FM_ERROR if the object is not a writable file.
 */
int filemapper_t::test_file_create(char *syspath)
{
	int exists = (probe(syspath) == FM_FILE);
	FILE *f = ::fopen(syspath, "a");
	if(f)
	{
		fclose(f);
		if(exists)
		{
			DBG(printf("  File iswritable!\n");)
			return FM_FILE;
		}
		else
		{
			DBG(printf("  File created!\n");)
			return FM_FILE_CREATE;
		}
	}
	else
	{
		if(errno == EISDIR)
		{
			DBG(printf("  Is a directory.\n");)
			return FM_DIR;
		}
		else
			DBG(printf("  Nope.\n"));
	}
	return FM_ERROR;
}


int filemapper_t::test_file_dir_any(char *syspath, int kind)
{
	int res = probe(syspath);
	switch (res)
	{
	  case FM_ERROR:
		DBG(printf("  Nope.\n");)
		break;
	  case FM_FILE:
		if((kind == FM_ANY) || (kind == FM_FILE))
		{
			DBG(printf("  Found! (File)\n");)
			return 1;
		}
		break;
	  case FM_DIR:
		if((kind == FM_ANY) || (kind == FM_DIR))
		{
			DBG(printf("  Found! (Dir)\n");)
			return 1;
		}
		break;
	  case FM_ANY:
		fprintf(stderr, "filemapper_t::probe()"
				" returned FM_ANY!?\n");
		break;
	}
	if(res != FM_ERROR)
		switch (kind)
		{
		  case FM_FILE:
			DBG(printf("  Not a file.\n");)
			break;
		  case FM_DIR:
			DBG(printf("  Not a directory.\n");)
			break;
		  default:
			break;
		}
	return 0;
}


char *filemapper_t::sys2unix(const char *syspath)
{
	char *buffer = salloc();
	strncpy(buffer, syspath, sizeof(buffer)-1);
	buffer[sizeof(buffer)-1] = 0;
	unix_slashes(buffer);
	return buffer;
}


char *filemapper_t::unix2sys(const char *path)
{
	char *buffer = salloc();
	strncpy(buffer, path, sizeof(buffer)-1);
	buffer[sizeof(buffer)-1] = 0;
	sys_slashes(buffer);
	return buffer;
}


/*
 * Try the specified operation on a path.
 */
int filemapper_t::try_get(const char *path, int kind)
{
	char buffer[FM_BUFFER_SIZE];
	strncpy(buffer, path, FM_BUFFER_SIZE);
	sys_slashes(buffer);
	switch (kind)
	{
	  case FM_FILE:
	  case FM_DIR:
	  case FM_ANY:
		if(test_file_dir_any(buffer, kind))
			return 1;
		break;
	  case FM_FILE_CREATE:
		switch (test_file_create(buffer))
		{
		  case FM_FILE:
		  case FM_FILE_CREATE:
			return 1;
		  default:
			break;
		}
		break;
	  default:
		break;
	}
	return 0;
}


/*
 * Parse the specified path, recurse through all possible
 * combinations, trying each terminal with try_get().
 */
int filemapper_t::recurse_get(char *result, const char *ref, int kind,
		int level = 1)
{
	DBG(printf(" [level %d]\n", level));
	char buffer[FM_BUFFER_SIZE];
	strncpy(buffer, ref, FM_BUFFER_SIZE);

	char *obj = strstr(ref, FM_DEREF_TOKEN);
	if(!obj)
		return 0;

	char kclass[9];
	int len = obj - ref;
	obj += 2;
	strncpy(kclass, ref, 8);
	if(len > 8)
		len = 8;
	kclass[len] = 0;

	fm_key_t *k = NULL;
	int hits = 0;
	while( (k = getkey(k, kclass)) )
	{
		++hits;
		snprintf(buffer, FM_BUFFER_SIZE, "%s/%s", k->path, obj);
		buffer[FM_BUFFER_SIZE-1] = 0;
		DBG({for(int s=0;s<level;++s)printf("  ");});
		int res;
		if(strstr(buffer, FM_DEREF_TOKEN))
		{
			DBG(printf("Reparsing '%s'", k->path));
			res = recurse_get(buffer, buffer, kind, level + 1);
		}
		else
		{
			DBG(printf("Trying '%s' -> '%s'...", k->key,
					k->path));
			res = try_get(buffer, kind);
		}
		if(res)
		{
			strncpy(result, buffer, FM_BUFFER_SIZE);
			return 1;
		}
	}
	DBG(	if(!hits)
		{
			for(int s=0;s<level;++s)printf("  ");
			printf("No paths registered for '%s'!\n", kclass);
		}
	);
	return 0;
}


char *filemapper_t::get(const char *ref, int kind = FM_FILE)
{
	DBG(
		if(kind == FM_FILE_CREATE)
			printf("Trying to create '%s'...", ref);
		else if(kind == FM_DIR)
			printf("Looking for '%s'...", ref);
		else
			printf("Looking for '%s'...", ref);
	);

	char *buffer = salloc();
	if(recurse_get(buffer, ref, kind))
		return buffer;
	else
		return NULL;
}


FILE *filemapper_t::fopen(const char *ref, const char *mode)
{
	DBG(fprintf(stderr, "filemapper_t::fopen()"));
	const char *path = NULL;
	int md = 0;

	if(strchr(mode, 'r'))
		md = 1;
	else if(strchr(mode, 'w'))
		md = 3;
	else if(strchr(mode, 'a'))
		md = 5;
	if(md)
		if(strchr(mode, '+'))
			md += 1;

	switch(md)
	{
	  case 0:	//Error!
		return NULL;
	  case 1:	//r
	  case 2:	//r+
		path = get(ref, FM_FILE);
		break;
	  case 3:	//w
	  case 4:	//w+  
	  case 5:	//a
	  case 6:	//a+
		path = get(ref, FM_FILE_CREATE);
		break;
	}
	if(path)
		return ::fopen(path, mode);
	else
		return NULL;
}


DIR *filemapper_t::opendir(const char *ref)
{
	DBG(fprintf(stderr, "filemapper_t::opendir()"));
	const char *path = get(ref, FM_DIR);
	if(path)
		return ::opendir(path);
	else
		return NULL;
}


int filemapper_t::mkdir(const char *ref, int perm)
{
	fprintf(stderr, "filemapper_t::mkdir not yet implemented!\n");
	return -1;
#if 0
	const char *path = get(ref, FM_DIR_CREATE);
//	::mkdir(path, perm);
#endif
}


void filemapper_t::print(FILE *f, const char *ref)
{
	char key[9];
	key[8] = 0;
	char *obj = strstr(ref, FM_DEREF_TOKEN);
	if(!obj)
		strncpy(key, ref, 8);
	else
	{
		int len = obj - ref;
		strncpy(key, ref, 8);
		if(len > 8)
			len = 8;
		key[len] = 0;
	}
	int all = (key[0] == '*');
	fm_key_t *k = keys;
	while(k)
	{
		if(strcmp(k->key, key) == 0 || all)
			fprintf(f, "%s"FM_DEREF_TOKEN" --> \"%s\"\n", k->key, k->path);
		k = k->next;
	}
}
