/*
 * Simple "GUI" toolkit for config screens.
 *  David Olofson, 2001
 */


#include "config.h"

#include <string.h>
#include <stdlib.h>

#include "window.h"
#include "toolkit.h"



/*----------------------------------------------------------
	ct_widget_t::
----------------------------------------------------------*/

ct_widget_t::ct_widget_t()
{
	interactive = 0;
	transparent = 1;
	highlighted = 0;
	user = NULL;
	tag = 0;
	_color = 0;
	next = prev = NULL;
	_widget_index = -1;
}


void ct_widget_t::init(gfxengine_t *e)
{
	window_t::init(e);
	_color = alloc_color(0,0,0);
}


void ct_widget_t::transparency(int t)
{
	if(t == transparent)
		return;

	transparent = t;
	if(!transparent)
		render();
}


void ct_widget_t::highlight(int hl)
{
	if(hl == highlighted)
		return;

	highlighted = hl;
	if(!transparent)
		render();
}


void ct_widget_t::color(Uint32 _cl)
{
	if(_cl == _color)
		return;

	_color = _cl;
	if(!transparent)
		render();
}


void ct_widget_t::render()
{
	if(!transparent)
	{
		foreground(_color);
		fillrect(0, 0, width(), height());
	}
	if(highlighted)
		ct_engine.render_highlight(this);
}


void ct_widget_t::change(int delta)
{
	_value += delta;
}


void ct_widget_t::value(int val)
{
	_value = val;
}


int ct_widget_t::value()
{
	return _value;
}



/*----------------------------------------------------------
	ct_engine_t::
----------------------------------------------------------*/

void default_render_highlight(ct_widget_t *wg)
{
	Uint32 hlc[6];
	hlc[0] = wg->alloc_color(0x4444,0,0);
	hlc[1] = wg->alloc_color(0x6666,0,0);
	hlc[2] = wg->alloc_color(0x8888,0,0);
	hlc[3] = wg->alloc_color(0x9999,0,0);
	hlc[4] = wg->alloc_color(0xaaaa,0,0);
	hlc[5] = wg->alloc_color(0xbbbb,0,0);
	wg->foreground(hlc[5]);
	wg->fillrect(0, 6, wg->width(), wg->height()-6-1);
	for(int i = 0; i < 6; ++i)
	{
		wg->foreground(hlc[i]);
		wg->fillrect(0, i, wg->width(), 1);
		wg->fillrect(0, wg->height()-1-i, wg->width(), 1);
	}
}


ct_engine_t::ct_engine_t()
{
	ct_engine.render_highlight = default_render_highlight;
}


ct_engine_t ct_engine;



/*----------------------------------------------------------
	ct_label_t::
----------------------------------------------------------*/

ct_label_t::ct_label_t(const char *cap = NULL)
{
	if(cap)
		caption(cap);
	else
		caption("Label");
}


void ct_label_t::caption(const char *cap)
{
	strncpy(_caption, cap, sizeof(_caption));
	_caption[sizeof(_caption)-1] = 0;
	if(!transparent)
		render();
}


void ct_label_t::render()
{
	ct_widget_t::render();
	center(0, _caption);
}



/*----------------------------------------------------------
	ct_item_t::
----------------------------------------------------------*/

ct_item_t::ct_item_t(const char *cap = NULL, int val = 0)
{
	if(cap)
		caption(cap);
	else
		caption("Item");
	_value = val;
}


void ct_item_t::caption(const char *cap)
{
	strncpy(_caption, cap, sizeof(_caption));
	_caption[sizeof(_caption)-1] = 0;
}


static ct_item_t dummy_item("<error>", -1);


/*----------------------------------------------------------
	ct_list_t::
----------------------------------------------------------*/

ct_list_t::ct_list_t(const char *cap = NULL) : ct_label_t(cap)
{
	interactive = 1;
	if(cap)
		caption(cap);
	else
		caption("List");
	items = NULL;
	_selected = NULL;
}


ct_list_t::~ct_list_t()
{
	clean();
}


void ct_list_t::clean()
{
	if(!items)
		return;

	/*
	 * Break the circular list first, or we'll
	 * blow up the memory manager...! :-)
	 */
	items->prev->next = NULL;

	while(items)
	{
		ct_item_t *i = items;
		items = i->next;
		delete i;
	}

	items = NULL;
	_selected = NULL;
}


void ct_list_t::add(ct_item_t *item)
{
	if(items)
	{
		item->next = items;
		item->prev = items->prev;
		items->prev->next = item;
		items->prev = item;
	}
	else
	{
		item->next = item->prev = item;
		items = item;
		_selected = item;
		if(!transparent)
			render();
	}
}


void ct_list_t::add(const char *cap, int val)
{
	ct_item_t *it = new ct_item_t(cap, val);
	add(it);
}


void ct_list_t::select(ct_item_t *item)
{
	_selected = item;
#if 0
	ct_item_t *i = items;
	if(i)
		for(int n = 0; n < 10000; ++n)	//Just in case...
		{
			if(i == _selected)
			{
				_value = n;
				break;
			}
			i = i->next;
		}
#endif
	if(_selected)
		_value = _selected->value();
	if(!transparent)
		render();
}


void ct_list_t::select(int index)
{
	if(!items)
		return;

	_selected = items;
	for(int i = 0; i < index; ++i)
		_selected = _selected->next;
	select(_selected);	//To update _value, render etc...
}


ct_item_t *ct_list_t::selected()
{
	if(!items)
		return &dummy_item;

	return _selected;
}


void ct_list_t::value(int val)
{
	if(!items)
		return;

	ct_item_t *best = items;
	_selected = items;
	while(_selected->value() != val)
	{
		if(labs(_selected->value()-val) < labs(best->value()-val))
			best = _selected;
		if(_selected->next == items)
			break;
		_selected = _selected->next;
	}
	if(_selected->value() != val)
		_selected = best;
	select(_selected);	//To update _value, render etc...
}


int ct_list_t::value()
{
	if(!items)
		return dummy_item.value();
	if(_selected)
		return _selected->value();
	else
		return dummy_item.value();
}


void ct_list_t::change(int delta)
{
	if(_selected)
	{
		if(delta > 0)
			for(int i = 0; i < delta; ++i)
				_selected = _selected->next;
		else if(delta < 0)
			for(int i = 0; i > delta; --i)
				_selected = _selected->prev;
		select(_selected);	//To update _value, render etc...
	}
	ct_widget_t::change(delta);
}


void ct_list_t::render()
{
	char buf[128];
	const char *ic;
	if(_selected)
		ic = _selected->caption();
	else
		ic = dummy_item.caption();
	ct_widget_t::render();
	snprintf(buf, sizeof(buf), "%s: %s", caption(), ic);
	center(0, buf);
}



/*----------------------------------------------------------
	ct_spin_t::
----------------------------------------------------------*/

ct_spin_t::ct_spin_t(const char *cap = NULL, int _min, int _max) : ct_label_t(cap)
{
	interactive = 1;
	min = _min;
	max = _max;
}


void ct_spin_t::value(int val)
{
	val -= min;
	while(val < 0)
		val += max - min + 1;
	val %= max - min + 1;
	val += min;
	ct_label_t::value(val);
}


void ct_spin_t::change(int delta)
{
	value(_value + delta);
}


void ct_spin_t::render()
{
	char buf[128];
	ct_widget_t::render();
	snprintf(buf, sizeof(buf), "%s: %d", caption(), _value);
	center(0, buf);
}



/*----------------------------------------------------------
	ct_button_t::
----------------------------------------------------------*/

ct_button_t::ct_button_t(const char *cap = NULL) : ct_label_t(cap)
{
	interactive = 1;
}


void ct_button_t::render()
{
	ct_label_t::render();
}


void ct_button_t::change(int delta)
{
	ct_label_t::change(delta);
	if(!transparent)
		render();
}



/*----------------------------------------------------------
	ct_form_t::
----------------------------------------------------------*/

ct_form_t::ct_form_t()
{
	widgets = NULL;
	_selected = NULL;
}


ct_form_t::~ct_form_t()
{
	if(!widgets)
		return;

	/*
	 * Break the circular list first, or we'll
	 * blow up the memory manager...! :-)
	 */
	widgets->prev->next = NULL;

	while(widgets)
	{
		ct_widget_t *i = widgets;
		widgets = i->next;
		delete i;
	}
}


void ct_form_t::add(ct_widget_t *widget)
{
	if(widgets)
	{
		widget->next = widgets;
		widget->prev = widgets->prev;
		widgets->prev->next = widget;
		widgets->prev = widget;
		widget->_widget_index = widget->prev->_widget_index + 1;
	}
	else
	{
		widget->_widget_index = 0;
		widget->next = widget->prev = widget;
		widgets = widget;
		select(widget);
	}
	render();
}


void ct_form_t::clean()
{
	if(!widgets)
		return;

	/*
	 * Break the circular list first, or we'll
	 * blow up the memory manager...! :-)
	 */
	widgets->prev->next = NULL;

	while(widgets)
	{
		ct_widget_t *w = widgets;
		widgets = w->next;
		delete w;
	}

	widgets = NULL;
	_selected = NULL;
}


void ct_form_t::select(ct_widget_t *widget)
{
	if(_selected)
		_selected->highlight(0);
	_selected = widget;
	if(_selected)
	{
		_selected->highlight(1);
		if(!_selected->transparent)
			_selected->render();
	}
}


void ct_form_t::select(int index)
{
	if(!widgets)
		return;

	if(_selected)
		_selected->highlight(0);
	_selected = widgets;
	for(int i = 0; i < index; ++i)
		_selected = _selected->next;
	if(_selected)
	{
		_selected->highlight(1);
		if(!_selected->transparent)
			_selected->render();
	}
}


void ct_form_t::next()
{
	if(!widgets)
		return;

	if(_selected)
		select(_selected->next);
}


void ct_form_t::prev()
{
	if(!widgets)
		return;

	if(_selected)
		select(_selected->prev);
}


ct_widget_t *ct_form_t::selected()
{
	return _selected;
}


int ct_form_t::selected_index()
{
	if(_selected)
		return _selected->_widget_index;
	else
		return -1;
}


void ct_form_t::change(int delta)
{
	if(_selected)
		_selected->change(delta);
}


void ct_form_t::render()
{
	ct_widget_t *w = widgets;
	while(w)
	{
		w->render();
		w = w->next;
		if(w == widgets)
			break;	//Done!
	}
}
