/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 1995, 1996, Akira Higuchi
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#if !defined(WIN32) && !defined(MACOS)
#include <pwd.h>
#endif

#include "kobo.h"
#include "score.h"
#include "file.h"


score_manager_t scorefile;


class uint_little_endian
{
	unsigned char byte[4];
  public:
	uint_little_endian(unsigned int x = 0);
	operator unsigned int ();
};

uint_little_endian::uint_little_endian(unsigned int x)
{
	byte[0] = x & 0xff;
	x >>= 8;
	byte[1] = x & 0xff;
	x >>= 8;
	byte[2] = x & 0xff;
	x >>= 8;
	byte[3] = x;
};

uint_little_endian::operator unsigned int ()
{
	return byte[0] | (byte[1] << 8) | (byte[2] << 16) | (byte[3] <<
			24);
};


struct s_entry_t
{
	uint_little_endian score;
	uint_little_endian scene;
	char name[64];
};

int s_table_cmp(s_entry_t * a, s_entry_t * b)
{
	if(a->score < b->score)
		return 1;
	else if(a->score > b->score)
		return -1;
	return 0;
}


score_manager_t::score_manager_t()
{
	_highscore = 0;
	_last_scene = 0;
	login_name[0] = 0;
}


score_manager_t::~score_manager_t()
{
}


void score_manager_t::init()
{
	const char *p;
	int i;
	s_entry_t tbl;
	memset(&tbl, 0, sizeof(tbl));
#ifdef WIN32
/*
FIXME: This should be properly implemented on Windoze - if we
FIXME: decide that it's actually much point. (Users, security etc
FIXME: are rather foreign matters to the average winuser...)
*/
	strcpy(login_name, "Unknown");
	p = fmap.get("SCORES>>42");
#else
	strcpy(tbl.name, "foo");
	struct passwd *pw = getpwuid(getuid());
	if(pw == NULL)
	{
		fprintf(stderr, "NOTE: getpwuid() failed;"
				" using UID for name.\n");
		snprintf(login_name, sizeof(login_name), "%u", getuid());
	}
	else
	{
		strncpy(login_name, pw->pw_name, sizeof(login_name));
		login_name[sizeof(login_name)-1] = 0;
	}
	{
		char buf[1024];
		snprintf(buf, sizeof(buf), "SCORES>>%u", getuid());
		p = fmap.get(buf);
	}
#endif
//	fprintf(stderr, "path = %s\n", save_path);
	if((i = load_block(p, &tbl, sizeof(tbl))) < 0)
		fprintf(stderr, "WARNING: Can't read score file!\n");
	_highscore = tbl.score;
	_last_scene = tbl.scene;
	show_high_scores(0);
}


void score_manager_t::player(const char *name)
{
	strncpy(login_name, name, sizeof(login_name));
	login_name[sizeof(login_name)-1] = 0;
}


void score_manager_t::record(int new_score, int stage)
{
	const char *p;
	int write = 0;
	s_entry_t tbl;
	memset(&tbl, 0, sizeof(tbl));
	strcpy(tbl.name, "foo");
#ifdef WIN32
	p = fmap.get("SCORES>>42");
#else
	{
		char buf[1024];
		snprintf(buf, sizeof(buf), "SCORES>>%u", getuid());
		p = fmap.get(buf);
	}
#endif
	(void)load_block(p, &tbl, sizeof(tbl));
	if((unsigned int)new_score > tbl.score)
	{
		tbl.score = new_score;
		write = 1;
	}
	if((unsigned int)stage > tbl.scene)
	{
		tbl.scene = stage;
		write = 1;
	}
	_highscore = tbl.score;
	_last_scene = tbl.scene;
	if(write)
	{
#ifdef WIN32
		p = fmap.get("SCORES>>42", FM_FILE_CREATE);
#else
		char buf[1024];
		snprintf(buf, sizeof(buf), "SCORES>>%u", getuid());
		p = fmap.get(buf, FM_FILE_CREATE);
#endif
		memset(tbl.name, 0, sizeof(tbl.name));
		strcpy(tbl.name, login_name);
		if(save_block(p, &tbl, sizeof(tbl)) < 0)
			fprintf(stderr, "WARNING: Can't update score file!\n");
		show_high_scores(0);

#ifdef HAS_HOME
		const char *bak_path = fmap.get("HOME>>kobodlsc.bak", FM_FILE_CREATE);
		if(bak_path)
		{
			if(save_block(bak_path, &tbl, sizeof(tbl)) < 0)
				fprintf(stderr, "WARNING: Can't write score file backup!\n");
		}
		else
			fprintf(stderr, "WARNING: No home dir for score file backup!\n");
#endif
	}
}


void score_manager_t::show_high_scores(int to_stdout)
{
/*
FIXME: No multiple windows support in SDL, so we'll have to rearrange this some...
*/
	to_stdout = 1;

	s_entry_t tbl[SCORE_SAVE_MAX];
	char path[1024];
	int i, j;
	for(i = 0; i < SCORE_SAVE_MAX; i++)
	{
		tbl[i].score = 0;
		tbl[i].scene = 0;
		strcpy(tbl[i].name, "foo");
	}

	struct dirent *direntp;
	DIR *dp;
	struct stat st_buffer;
	const char *dirpath = fmap.get("SCORES>>", FM_DIR);
	dp = fmap.opendir("SCORES>>");
	if(dp == NULL)
		return;
	i = 0;
	while((direntp = readdir(dp)) != NULL)
	{
		sprintf(path, "%s%s", dirpath, direntp->d_name);
		if(stat(path, &st_buffer))
			continue;
		if(!S_ISREG(st_buffer.st_mode))
			continue;
		if(load_block(path, &tbl[i], sizeof(s_entry_t)) < 0)
			continue;
		if((++i) >= SCORE_SAVE_MAX)
			break;
	}
	closedir(dp);

	qsort(tbl, i, sizeof(s_entry_t),
			(int (*)(const void *, const void *))s_table_cmp);

	if(to_stdout)
	{
		printf("Name                     Score    Stages\n");
		for(j = 0; j < i; j++)
		{
			printf("%-20s %9u %9u\n", tbl[j].name,
					(unsigned int)tbl[j].score,
					(unsigned int)tbl[j].scene);
		}
	}
	else
	{
#if 0
		wscore.clear();
		char buffer[256];
		if(i > SCORE_SHOW_MAX)
			i = SCORE_SHOW_MAX;
		wscore.string(10, 24, "name");
		wscore.string(180, 24, "high score");
		wscore.string(80, 24, "cleared stages");
		for(j = 0; j < i; j++)
		{
			wscore.string(10, 16 * j + 48, tbl[j].name);
			sprintf(buffer, "%09u",
					(unsigned int)tbl[j].score);
			wscore.string(180, 16 * j + 48, buffer);
			sprintf(buffer, "%09u",
					(unsigned int)tbl[j].scene);
			wscore.string(100, 16 * j + 48, buffer);
		}
#endif
	}
}
