/*
 * Written for SGI DMedia API by Masanao Izumo <mo@goice.co.jp>
 * Ported to OSS by David Olofson <do@reologica.se>, 2001
 *
FIXME: The command FIFO I added isn't timestamped.
FIXME: /David
 *
 * TODO:
 *	* Figure out what causes audio to behave
 *	  weird on some machines...
 *
 *	* Improve resampling quality.
 *
 *	* Add mmap() support for lowlatency sfx.
 */

#include "config.h"

#define	DBG2(x)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef OSS_SUPPORT
#include <pthread.h>
#include <unistd.h>
#include "audiodev.h"
#endif

#include "SDL.h"
#include "SDL_audio.h"

#include "audio.h"
#include "sfifo.h"

#define BUFFERS		3
#define	MAX_COMMANDS	64

#define	OSCFRAMES	240
DBG(
int *oscbuf = NULL;
int oscframes = 0;
static int oscpos = 0;
);

int samplerate = 44100;
int lim_threshold = DEFAULT_LIM_THRESHOLD;
int lim_release = DEFAULT_LIM_RELEASE;
int attenuation = 0;

int default_ctl[AC_COUNT] = {
	0,	//AC_PLAYRESET
	0,	//AC_LOOP
	0,	//AC_RETRIG
	0,	//AC_RANDTRIG

	100,	//AC_VOLUME	[100%]
	65536,	//AC_SPEED	[1.0 * output_sample_rate]
	65536,	//AC_PITCH	[1.0 * AC_SPEED_sample_rate]
};


/*----------------------------------------------------------
	Asynchronous command interface stuff
----------------------------------------------------------*/

typedef struct
{
	enum
	{
		CMD_STOP = 0,
		CMD_PLAY,
		CMD_CONTROL
	} action;
	int	cid;
	union
	{
		struct
		{
			char	*start;
			int	length;
		} sample;
		struct
		{
			int	ctl;
			int	arg;
		} control;
	} args;
} command_t;

static sfifo_t commands;


/*--------------------------------------------------------------------
	delay_t - Feedback delay w/ LP filter
--------------------------------------------------------------------*/

#define	DELAY_BUFSIZE	32768

typedef struct dtap_t
{
	int	delay;
	int	shift;
} dtap_t;

dtap_t tapinit[] = {
	//Filtered feedback taps
	{24350, 2},	{24360, 2},
	{34560, 2},	{37550, 2},
	{53100, 2},	{53200, 2},

	//Normal taps
	{0, 0},		{0, 0},
	{3000, 3},	{5000, 2},
	{7000, 4},	{8000, 5},
	{11000, 5},	{13000, 6},
};

#define	MAX_DELAY_TAPS	16
#define	DELAY_TAPS	(sizeof(tapinit) / sizeof(dtap_t))

typedef struct delay_t
{
	int	*buf;
	int	*loopbuf;
	int	inspos;
	int	cl;
	int	cr;
	dtap_t	taps[MAX_DELAY_TAPS];
} delay_t;


/*----------------------------------------------------------
	Audio channel
----------------------------------------------------------*/

typedef Sint8 audio_samp_t;
typedef Sint16 out_samp_t;

typedef struct
{
	audio_samp_t *samp;
	int	nsamp;		/* number of samples */
	int	position_fxp;	/* 24:8 fixpoint */
	int	ctl[AC_COUNT];
} audio_channel_t;

static audio_channel_t audio_channel[MAX_CHANNEL];


/*----------------------------------------------------------
	Engine stuff
----------------------------------------------------------*/

#ifdef OSS_SUPPORT
static audiodev_t adev;
static pthread_t engine_thread;
#endif

static int using_oss = 0;

static SDL_AudioSpec audiospec;

static int running = 0;
static int master_vol = 100<<2;
static int master_rvb = 100<<2;

/*
 * Messy. bufsize is in mono samples, while we're
 * actually dealing with stereo here.
 */
static int *mixbuf = NULL;
static int bufsize = 64;

delay_t	reverb;


/*--------------------------------------------------------------------
	delay_t - Feedback delay w/ LP filter
--------------------------------------------------------------------*/

#define	DELAY_MASK	(DELAY_BUFSIZE-1)

int delay_init(delay_t *d, int samplerate)
{
	float scale;
	int i;

	d->inspos = 0;
	d->cl = 0;
	d->cr = 0;
	d->buf = calloc(1, sizeof(int)*DELAY_BUFSIZE);
	if(!d->buf)
		return -1;

	d->loopbuf = calloc(1, sizeof(int)*DELAY_BUFSIZE);
	if(!d->loopbuf)
		return -2;

	scale = (float)samplerate / 44100.0;
	for(i = 0; i < DELAY_TAPS; ++i)
	{
		d->taps[i].delay = (int)(tapinit[i].delay * scale) & (DELAY_MASK-1);
		d->taps[i].delay |= i & 1;
		d->taps[i].shift = tapinit[i].shift;
	}

	return 0;
}


void delay_close(delay_t *d)
{
	free(d->loopbuf);
	free(d->buf);
}


#define	DB(x)	delaybuf[(inspos-(x)) & DELAY_MASK]
#define	DTAP(n)	(DB(d->taps[n].delay) >> d->taps[n].shift)
#define	LB(x)	loopbuf[(inspos-(x)) & DELAY_MASK]
#define	TAP(n)	(LB(d->taps[n].delay) >> d->taps[n].shift)
void delay_process_mix_s(delay_t *d, int *buf, int frames)
{
	int *delaybuf = d->buf;
	int *loopbuf = d->loopbuf;
	int inspos = d->inspos;
	int cl = d->cl;
	int cr = d->cr;
	int s = frames;
	int outl, outr;
	int vm = (double)master_rvb / (double)DEFAULT_RVB;
	int i;
	while(s--)
	{
		/* Input */
		DB(0) = buf[0];
		DB(1) = buf[1];

		/* Feedback with LP filters */
		cl += (TAP(0) + TAP(3) + TAP(4) - cl) >> 3;
		cr += (TAP(1) + TAP(2) + TAP(5) - cr) >> 3;
		LB(0) = cl;
		LB(1) = cr;

		/* Taps */
		outl = outr = 0;
		for(i = 6; i < DELAY_TAPS; i += 2)
		{
			outl += TAP(i);
			outr += TAP(i+1);
		}

		LB(0) += DB(0);
		LB(1) += DB(0);

		*buf++ += (outl*vm)>>8;
		*buf++ += (outr*vm)>>8;
		inspos += 2;
	}
	d->cl = cl;
	d->cr = cr;
	d->inspos = inspos;
}
#undef	DB
#undef	DTAP
#undef	LB
#undef	TAP


/*----------------------------------------------------------
	Engine code
----------------------------------------------------------*/

static void __play_stop(int cid)
{
	audio_channel_t *c;
	c = audio_channel + cid;
	c->samp = NULL;
}


static void __play(int cid, char *samp, int nsamp)
{
	audio_channel_t *c;
	c = audio_channel + cid;
	c->samp = (audio_samp_t *) samp;
	c->nsamp = nsamp;
	c->position_fxp = 0;
	if(c->ctl[AC_PLAYRESET])
		memcpy(c->ctl+1, default_ctl+1, (AC_COUNT-1) * sizeof(int));
}


static void __run_commands(void)
{
	while(sfifo_used(&commands) >= sizeof(command_t))
	{
		command_t cmd;
		if(sfifo_read(&commands, &cmd, sizeof(cmd)) < 0)
		{
			fprintf(stderr, "audio.c: Engine failure!\n");
			running = 0;
			return;
		}
		switch(cmd.action)
		{
		  case CMD_STOP:
			DBG2(printf("Got CMD_STOP.\n"));
			__play_stop(cmd.cid);
			break;
		  case CMD_PLAY:
			DBG2(printf("Got CMD_PLAY.\n"));
			__play(cmd.cid, cmd.args.sample.start,
					cmd.args.sample.length);
			break;
		  case CMD_CONTROL:
			DBG2(printf("Got CMD_CONTROL.\n"));
			{
				audio_channel_t *c;
				c = audio_channel + cmd.cid;
				c->ctl[cmd.args.control.ctl] =
						cmd.args.control.arg;
			}
			break;
		}

	}
}


static void __push_command(command_t *cmd)
{
	if(!running)
		return;

	if(sfifo_space(&commands) < sizeof(command_t))
	{
		fprintf(stderr, "WARNING: Audio command FIFO overflow!\n");
		return;
	}
	sfifo_write(&commands, cmd, sizeof(command_t));
}


static void __mix_channel(audio_channel_t *c, int *buf, int frames)
{
	if(!c->samp)
		return;	/* Done --> */

	/* Loop until buffer full or end-of-waveform */
	while(frames > 0)
	{
		int ipos = (c->position_fxp >> 8) & 0xfffffffe;

		/* Mix from 8 bit stereo source */
		if(ipos < c->nsamp)
		{
			int j, n, v, sp, step;
			audio_samp_t *s;

			n = c->nsamp - ipos;
			v = c->ctl[AC_VOLUME];
			if(n > frames)
				n = frames;
			s = c->samp;
			sp = c->position_fxp;
			step = c->ctl[AC_SPEED] >> 7;	// Stereo!
			step *= c->ctl[AC_PITCH] >> 8;
			step >>= 8;
			if(v == DEFAULT_VOL)
			{
				for(j = 0; j < n; j += 2)
				{
					int ind = (sp >> 8) & 0xfffffffe;
					buf[j] += s[ind] << 8;
					buf[j + 1] += s[ind + 1] << 8;
					sp += step;
				}
			}
			else
			{
				int vm = (double)v / (double)DEFAULT_VOL;
				for(j = 0; j < n; j += 2)
				{
					int ind = (sp >> 8) & 0xfffffffe;
					buf[j] += s[ind] * vm;
					buf[j + 1] += s[ind + 1] * vm;
					sp += step;
				}
			}
			c->position_fxp = sp;
			ipos = (sp >> 8) & 0xfffffffe;
			frames -= n;
			buf += n;
		}

		/* Handle end-of-sample and loop */
		if(ipos >= c->nsamp)
		{
			if(c->ctl[AC_LOOP])
			{
				c->position_fxp &= 0xff; /* Keep decimals! */
				DBG2(printf("Sound looped.\n"));
			}
			else
			{
				c->samp = NULL;
				DBG2(printf("Sound ended.\n"));
				return;
			}
		}
	}
}


static void __mix(int *buf, int frames)
{
	int i;
	memset(buf, 0, frames * sizeof(int));
	for (i = 0; i < MAX_CHANNEL; i++)
		__mix_channel(audio_channel + i, buf, frames);
}


/*
 * Apply master volume
 */
static void __mastervol(int *buf, int frames)
{
	int i;
	for(i = 0; i < frames; ++i)
	{
		int s = buf[i];
		s *= master_vol;
		s >>= 8;
		buf[i] = s;
	}
}


#define	CLIP_MIN	-32700
#define	CLIP_MAX	32700

/*
 * Convert with saturation
 */
static void __clip(int *inbuf, out_samp_t *outbuf, int frames)
{
	int i;
	for(i = 0; i < frames; ++i)
	{
		int s = inbuf[i];
		s *= master_vol;
		s >>= 8;
		if(s < CLIP_MIN)
			outbuf[i] = CLIP_MIN;
		else if(s > CLIP_MAX)
			outbuf[i] = CLIP_MAX;
		else
			outbuf[i] = s;
	}
}


static void __limiter(int *buf, int frames)
{
	static int peak = 32768;
	int i;
	for(i = 0; i < frames; ++i)
	{
		int s = buf[i];
		if(s > peak)
			peak = s;
		else if(-s > peak)
			peak = -s;
		else
		{
			peak -= lim_release;
			if(peak < lim_threshold)
				peak = lim_threshold;
		}

		buf[i] = (s << 15) / peak;
	}
	attenuation = peak - lim_threshold;
}


DBG(
static void __grab(out_samp_t *buf, int frames)
{
	int i;
	frames /= 2;
	for(i = 0; i < frames; ++i)
	{
		oscbuf[oscpos] = buf[i<<1];
		oscbuf[oscpos] += buf[(i<<1)+1];
		++oscpos;
		if(oscpos >= oscframes)
			oscpos = 0;
	}
}
);

/*
 * Engine callback for SDL_audio and OSS
 */
static void __audio_callback(void *ud, Uint8 *stream, int len)
{
	out_samp_t *outbuf = (out_samp_t *)stream;
	int frames = len / sizeof(out_samp_t);
	__run_commands();
	__mix(mixbuf, frames);
	if(master_rvb)
		delay_process_mix_s(&reverb, mixbuf, frames/2);
	__mastervol(mixbuf, frames);
	__limiter(mixbuf, frames);
	__clip(mixbuf, outbuf, frames);
	DBG(__grab(outbuf, frames));
}


#ifdef OSS_SUPPORT
/*
 * Engine thread for OSS
 */
void *__audio_engine(void *dummy)
{
	int bytes = bufsize * sizeof(out_samp_t);
	out_samp_t *outbuf = calloc(1, bytes);
	if(!outbuf)
	{
		audiodev_close(&adev);
		running = 0;
		return NULL;
	}
	while(running)
	{
		__audio_callback(NULL, (Uint8 *)outbuf, bytes);
		write(adev.outfd, outbuf, bytes);
	}
	audiodev_close(&adev);
	return NULL;
}
#endif


static int __init_oss_audio()
{
#ifdef OSS_SUPPORT
	mixbuf = calloc(1, bufsize * sizeof(int));
	if(!mixbuf)
		return -1;

	audiodev_init(&adev);
	adev.rate = samplerate;
	adev.fragmentsize = bufsize * sizeof(out_samp_t);
	adev.fragments = BUFFERS;
	running = 1;
	if(audiodev_open(&adev) < 0)
	{
		sfifo_close(&commands);
		fprintf(stderr, "audio.c: Failed to open audio!\n");
		running = 0;
		return -2;
	}

	if(pthread_create(&engine_thread, NULL, __audio_engine, NULL))
	{
		fprintf(stderr, "audio.c: Failed to start audio engine!\n");
		return -3;
	}
	return 0;
#else
	fprintf(stderr, "OSS audio not compiled in!\n");
	return -1;
#endif
}


static int __init_SDL_audio()
{
	SDL_AudioSpec as;

	if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0)
		return -2;

	as.freq = samplerate;
	as.format = AUDIO_S16SYS;
	as.channels = 2;
	as.samples = bufsize / as.channels;
	as.callback = __audio_callback;
	if(SDL_OpenAudio(&as, &audiospec) < 0)
		return -3;

	if(audiospec.format != AUDIO_S16SYS)
	{
		SDL_CloseAudio();
		return -4;
	}

	mixbuf = calloc(1, audiospec.samples *
			sizeof(int) * audiospec.channels);
	if(!mixbuf)
	{
		SDL_CloseAudio();
		return -1;
	}

	running = 1;
	SDL_PauseAudio(0);
	return 0;
}


/*----------------------------------------------------------
	Open/close code
----------------------------------------------------------*/

int audio_open(int rate, int latency, int use_oss)
{
	int i;
	running = 0;

	samplerate = rate;
	using_oss = use_oss;

	if(sfifo_init(&commands, sizeof(command_t) * MAX_COMMANDS) < 0)
	{
		fprintf(stderr, "audio.c: Failed to set up audio engine!\n");
		return -1;
	}

	memset(audio_channel, 0, sizeof(audio_channel));

	audio_default_control(AC_SPEED, 8000*65536/samplerate);

	for (i = 0; i < MAX_CHANNEL; i++)
		memcpy(&audio_channel[i].ctl, &default_ctl,
				sizeof(default_ctl));

	/*
	 * Note: We assume here that SDL also uses 3
	 *       fragments, which may not be true.
	 */
	if(latency > 1000)
		latency = 1000;
	bufsize = 64;
	while(BUFFERS*bufsize*1000 / (samplerate*2) < latency)
		bufsize <<= 1;
	bufsize >>= 1;

	if(delay_init(&reverb, samplerate) < 0)
		return -2;

	DBG(
		oscbuf = calloc(1, OSCFRAMES*sizeof(int));
		oscframes = OSCFRAMES;
	);

	if(using_oss)
		return __init_oss_audio();
	else
		return __init_SDL_audio();
}


void audio_close(void)
{
	if(!running)
		return;

	printf("Stopping audio engine... ");
	running = 0;
	if(using_oss)
#ifdef OSS_SUPPORT
		pthread_join(engine_thread, NULL);
#else
		;
#endif
	else
		SDL_CloseAudio();
	delay_close(&reverb);
	DBG(
		oscframes = 0;
		free(oscbuf);
		oscbuf = NULL;
	);
	printf("Ok.\n");
}


/*----------------------------------------------------------
	Channel control API code
----------------------------------------------------------*/

void audio_master_volume(int vol)
{
	master_vol = vol << 2;
}


void audio_master_reverb(int rvb)
{
	master_rvb = rvb << 8;
}


void audio_set_limiter(float thres, float rels)
{
	lim_threshold = (int)(32768.0 * thres);
	if(lim_threshold < 256)
		lim_threshold = 256;
	lim_release = (int)(rels * 65536.0 / samplerate);
}


void audio_volume(int cid, int vol)
{
	command_t cmd;
	cmd.action = CMD_CONTROL;
	cmd.cid = cid;
	cmd.args.control.ctl = AC_VOLUME;
	cmd.args.control.arg = vol;
	__push_command(&cmd);
}

void audio_play(int cid, char* samp, int nsamp)
{
	command_t cmd;
	cmd.action = CMD_PLAY;
	cmd.cid = cid;
	cmd.args.sample.start = samp;
	cmd.args.sample.length = nsamp;
	__push_command(&cmd);
}

void audio_control(int cid, int ctl, int arg)
{
	command_t cmd;
	cmd.action = CMD_CONTROL;
	cmd.cid = cid;
	cmd.args.control.ctl = ctl;
	cmd.args.control.arg = arg;
	__push_command(&cmd);
}

void audio_default_control(int ctl, int arg)
{
	int i;
	default_ctl[ctl] = arg;

	/* Apply to all silent channels */
	for(i = 0; i < MAX_CHANNEL; i++)
	{
		audio_channel_t *c = audio_channel + i;
		if(!c->samp)
			c->ctl[ctl] = arg;
	}

}

void audio_stop(int cid)
{
	command_t cmd;
	if(cid < 0)
		return;

	cmd.action = CMD_STOP;
	cmd.cid = cid;
	__push_command(&cmd);
}


/*
 * audio_is_active() bypasses the lock-free FIFO.
 * OTOH, it's meant to be asynchronous (I think...),
 * so it should do the right thing, provided the
 * reads are atomic.
 *		/David
 */
int audio_is_active(int cid)
{
	audio_channel_t *c;
	if(!running)
		return 0;

	c = audio_channel + cid;
	return c->samp != NULL;
}
